const fs=require('fs-extra')
const path = require('path')
const chalk = require('chalk')
const pkg = require(`${process.cwd()}/package.json`)

const defaults = {
  clean: true,
  target: 'app',
  formats: 'commonjs,umd,umd-min',
  'unsafe-inline': true
}

const buildModes = {
  lib: 'library',
  wc: 'web component',
  'wc-async': 'web component (async)'
}
if(!fs.existsSync(path.resolve('./src/pages'))){
  fs.mkdirSync(path.resolve('./src/pages'))
}

const commonPath = pkg.commonPath || (pkg.vueCliBuildConfig && pkg.vueCliBuildConfig.commonPath)
const excludePages = (pkg.vueCliBuildConfig && pkg.vueCliBuildConfig.exclude)? pkg.vueCliBuildConfig.exclude: []

const allPages = []
const dirs = fs.readdirSync(path.resolve('./src/pages/')).filter(page=>(page!=='.DS_Store' && excludePages.indexOf(page)<0))
for (let index = 0; index < dirs.length; index++) {
  const dirName = dirs[index];
  if(fs.existsSync(`${process.cwd()}/src/pages/${dirName}/index.ts`) && dirName !== 'common'){
    allPages.push(dirName)
  }
}
const modifyConfig = (config, fn) => {
  if (Array.isArray(config)) {
    config.forEach(c => fn(c))
  } else {
    fn(config)
  }
}
  
const _packageJson = fs.readFileSync(`${process.cwd()}/package.json`)
const { commonPath } = JSON.parse(_packageJson)

module.exports = (api, options) => {
  api.registerCommand('build', {
    description: 'build for production',
    usage: 'vue-cli-service build [options] [entry|pattern]',
    options: {
      '--mode': `specify env mode (default: production)`,
      '--dest': `specify output directory (default: ${options.outputDir})`,
      '--modern': `build app targeting modern browsers with auto fallback`,
      '--no-unsafe-inline': `build app without introducing inline scripts`,
      '--target': `app | lib | wc | wc-async (default: ${defaults.target})`,
      '--formats': `list of output formats for library builds (default: ${defaults.formats})`,
      '--name': `name for lib or web-component mode (default: "name" in package.json or entry filename)`,
      '--filename': `file name for output, only usable for 'lib' target (default: value of --name)`,
      '--no-clean': `do not remove the dist directory before building the project`,
      '--report': `generate report.html to help analyze bundle content`,
      '--report-json': 'generate report.json to help analyze bundle content',
      '--watch': `watch for changes`
    }
  }, async (args, rawArgs) => {
    for (const key in defaults) {
      if (args[key] == null) {
        args[key] = defaults[key]
      }
    }
    
    if(!args._[0]){
      if(allPages){
        args._[0] = allPages.join('+')
      } else {
        console.log(chalk.red('找不到需要打包的目录pages'))
        return
      }
    }
    
    const pageData =  getPages(args)
    if(pageData.length===0){
      console.log(chalk.red(`你所编译的页面不存在`))
      return 
    }
    pageData.forEach(async item=>{
      args.entry = item
      if (args.target !== 'app') {
        args.entry = args.entry || 'src/App.vue'
      }
  
      process.env.VUE_CLI_BUILD_TARGET = args.target
      if (args.modern && args.target === 'app') {
        process.env.VUE_CLI_MODERN_MODE = true
        delete process.env.VUE_CLI_MODERN_BUILD
        // if (!process.env.VUE_CLI_MODERN_BUILD) {
          // main-process for legacy build
          await build(Object.assign({}, args, {
            modernBuild: false,
            keepAlive: true
          }), api, options)
          // spawn sub-process of self for modern build
          // const { execa } = require('@vue/cli-shared-utils')
          // const cliBin = require('@vue/cli-service/bin/vue-cli-service.js')
          // await execa(cliBin, ['build', ...rawArgs], {
          //   stdio: 'inherit',
          //   env: {
          //     VUE_CLI_MODERN_BUILD: true
          //   }
          // })
        // } else {
          process.env.VUE_CLI_MODERN_BUILD = true
          // sub-process for modern build
          await build(Object.assign({}, args, {
            modernBuild: true,
            clean: false
          }), api, options)
        // }
        delete process.env.VUE_CLI_MODERN_MODE
        delete process.env.VUE_CLI_MODERN_BUILD
      } else {
        if (args.modern) {
          const { warn } = require('@vue/cli-shared-utils')
          warn(
            `Modern mode only works with default target (app). ` +
            `For libraries or web components, use the browserslist ` +
            `config to specify target browsers.`
          )
        }
        await build(args, api, options)
      }
      delete process.env.VUE_CLI_BUILD_TARGET
    })
  })
}

/**
 * 获取页面
 * @param {Array} args 
 */
function getPages(args) {
  let pages = []
  if(args._[0]){
   args._[0].split('+').forEach(item=>{
      if(allPages.indexOf(item)>=0){
        pages.push(item)
      }
   })
  }
  return pages
}

async function build (args, api, options) {
  const chalk = require('chalk')
  const webpack = require('webpack')
  const formatStats = require('../util/formatStats')
  const validateWebpackConfig = require('../util/validateWebpackConfig')
  const {
    log,
    done,
    info,
    logWithSpinner,
    stopSpinner
  } = require('@vue/cli-shared-utils')

  log()
  const mode = api.service.mode
  if (args.target === 'app') {
    const bundleTag = args.modern
      ? args.modernBuild
        ? `modern bundle `
        : `legacy bundle `
      : ``
    logWithSpinner(`Building ${args.entry} ${bundleTag}for ${mode}...`)
  } else {
    const buildMode = buildModes[args.target]
    if (buildMode) {
      const additionalParams = buildMode === 'library' ? ` (${args.formats})` : ``
      logWithSpinner(`Building for ${mode} as ${buildMode}${additionalParams}...`)
    } else {
      throw new Error(`Unknown build target: ${args.target}`)
    }
  }

  const targetDir = api.resolve(args.dest || options.outputDir) + '/' + args.entry
  const isLegacyBuild = args.target === 'app' && args.modern && !args.modernBuild

  // resolve raw webpack config
  let webpackConfig
  if (args.target === 'lib') {
    webpackConfig = require('../util/resolveLibConfig')(api, args, options)
  } else if (
    args.target === 'wc' ||
    args.target === 'wc-async'
  ) {
    webpackConfig = require('../util/resolveWcConfig')(api, args, options)
  } else {
    webpackConfig = require('../util/resolveAppConfig')(api, args, options)
  }

  // check for common config errors
  validateWebpackConfig(webpackConfig, api, options, args.target)
  
  // 获取页眉页脚
  const headerPath = path.resolve('./src/pages/common/header.html')
  const footerPath =  path.resolve('./src/pages/common/footer.html')
  const headerTpl = fs.existsSync(headerPath)?fs.readFileSync(headerPath):''
  const footerTpl =  fs.existsSync(footerPath)?fs.readFileSync(footerPath):''

  const entry = args.entry
  if (entry) {
    const pagePath = "./src/pages/" + entry
    webpackConfig.entry={}
    webpackConfig.entry[entry] = pagePath + '/index.ts'
    const curPublicPath = webpackConfig.output.publicPath
    webpackConfig.output.publicPath = ''
    webpackConfig.output.path = targetDir
    for(let item of webpackConfig.plugins) {
      if (item.options && item.options.template) {
        item.options.template = pagePath + '/index.html'
        item.options.header = headerTpl
        item.options.footer = footerTpl
        item.options.common = (commonPath && process.env.NODE_ENV === 'production')?`<link href=${commonPath} rel=preload as=script><script src="${commonPath}"></script>`:''
      }
    }
    for (let item of webpackConfig.module.rules) {
      if(item.test.toString().indexOf('png|jp')>=0){
        for (const itemChild of item.use) {
          if(itemChild.options){
            itemChild.options.publicPath = curPublicPath?curPublicPath.indexOf(entry)<0?curPublicPath+entry:curPublicPath:'/'+entry
            break
          }
        }
        break
      }
    }
  }

  // apply inline dest path after user configureWebpack hooks
  // so it takes higher priority
  if (args.dest) {
    modifyConfig(webpackConfig, config => {
      config.output.path = targetDir
    })
  }

  if (args.watch) {
    modifyConfig(webpackConfig, config => {
      config.watch = true
    })
  }

  // Expose advanced stats
  if (args.dashboard) {
    const DashboardPlugin = require('@vue/cli-service/lib/webpack/DashboardPlugin')
    modifyConfig(webpackConfig, config => {
      config.plugins.push(new DashboardPlugin({
        type: 'build',
        modernBuild: args.modernBuild,
        keepAlive: args.keepAlive
      }))
    })
  }

  if (args.report || args['report-json']) {
    const { BundleAnalyzerPlugin } = require('webpack-bundle-analyzer')
    modifyConfig(webpackConfig, config => {
      const bundleName = args.target !== 'app'
        ? config.output.filename.replace(/\.js$/, '-')
        : isLegacyBuild ? 'legacy-' : ''
      config.plugins.push(new BundleAnalyzerPlugin({
        logLevel: 'warn',
        openAnalyzer: false,
        analyzerMode: args.report ? 'static' : 'disabled',
        reportFilename: `${bundleName}report.html`,
        statsFilename: `${bundleName}report.json`,
        generateStatsFile: !!args['report-json']
      }))
    })
  }

  if (args.clean) {
    await fs.remove(api.resolve(args.dest || options.outputDir))
  }

  return new Promise((resolve, reject) => {
    webpack(webpackConfig, (err, stats) => {
      stopSpinner(false)
      if (err) {
        return reject(err)
      }

      if (stats.hasErrors()) {
        return reject(`Build failed with errors.`)
      }

      if (!args.silent) {
        const targetDirShort = path.relative(
          api.service.context,
          targetDir
        )
        log(formatStats(stats, targetDirShort, api))
        if (args.target === 'app' && !isLegacyBuild) {
          if (!args.watch) {
            done(`Build complete. The ${chalk.cyan(targetDirShort)} directory is ready to be deployed.`)
            info(`Check out deployment instructions at ${chalk.cyan(`https://cli.vuejs.org/guide/deployment.html`)}\n`)
          } else {
            done(`Build complete. Watching for changes...`)
          }
        }
      }

      // test-only signal
      if (process.env.VUE_CLI_TEST) {
        console.log('Build complete.')
      }

      resolve()
    })
  })
}

module.exports.defaultModes = {
  build: 'production'
}
